from django.db import models

from panya.models import ModelBase

class AgeRestriction(models.Model):
    """
    Simple age restriction model.
    """
    age = models.IntegerField(
        help_text="Rating age, i.e. 13, 18 etc.",
    )
    symbol = models.CharField(
        max_length=128,
        help_text='Rating symbol, i.e. PG, SN, R, G etc.',
        blank=True,
        null=True,
    )
    description = models.TextField(
        help_text='A short description. Usually an explanation of the symbol, i.e. Parental Guidance Suggested. Some material may not be suitable for children.',
        blank=True,
        null=True,
    )

class Channel(ModelBase):
    """
    Basic channel model.
    """
    age_restriction = models.ForeignKey(
        'tube.AgeRestriction',
        blank=True,
        null=True,
    )

class ClipBase(ModelBase):
    """
    Model defining clip metadata. Clip types should inherit from this model.
    I.e. Dexter - Dex Lies and Videotape.
    """
    duration = models.IntegerField(
        help_text="Duration of this clip in seconds. This is automatically determined from the media resource.",
        blank=True,
        null=True,
    )
    age_restriction = models.ForeignKey(
        'tube.AgeRestriction',
        blank=True,
        null=True,
    )
    channel = models.ForeignKey(
        'tube.Channel',
        help_text="Select a channel for which this clip should be available. This field is only honoured when a series is not specified. If a series is specified its channel takes priority.",
        blank=True,
        null=True,
    )
    series = models.ForeignKey(
        'tube.Series',
        help_text="Select a series for which this clip is an episode.",
        blank=True,
        null=True,
    )
    
    def get_channel(self):
        """
        Get channel for this clip.
        Return the series' channel if series is populated, otherwise channel.
        If no series or channel is populated return None.
        """
        # Return show's channel if this clip has a show associated with it.
        if self.series:
            if self.series.channel:
                return self.series.channel
        
        # Otherwise, return channel if this clip has a channel associated with it.
        if self.channel:
            return self.channel
       
        return None
    
    def get_channel_id(self):
        channel = self.get_channel()
        return channel.content_id if channel else None
            
    def get_series_title(self):
        return self.series.title if self.series else ''
        
    def get_series_id(self):
        return self.series.id if self.series else ''

    @property
    def duration_as_hours_minute_seconds(self):
        return {'hours': self.duration/60/60, 'minutes': (self.duration % 3600) / 60, 'seconds': self.duration % 60}
   
class MediaResourceAbstractClip(models.Model):
    """
    Abstract clip model with local media resource.
    """
    media_resource = models.FileField(
        help_text="Upload the media file for this clip.",
        upload_to='tube_clip_resources',
        blank=True,
        null=True,
    )

    class Meta():
        abstract = True

class Clip(ClipBase, MediaResourceAbstractClip):
    """
    Concrete Clip model with local media resource.
    """
    pass

class Episode(models.Model):
    """
    Model storing metadata associated with episode.
    I.e. Dexter - Dex Lies and Videotape, Season 2 Episode 6.
    """
    clip = models.ForeignKey(
        'tube.ClipBase',
    )
    season = models.ForeignKey(
        'tube.Season',
    )
    episode_number = models.IntegerField(
        help_text="Episode number of this clip.",
        blank=True,
        null=True,
    )
    
class Season(models.Model):
    """
    Model storing metadata associated with season.
    I.e. Simpsons Season 2.
    """
    clips = models.ManyToManyField(
        'tube.ClipBase',
        through='tube.Episode',
        blank=True,
        null=True,
    )

class Series(ModelBase):
    """
    Model storing metadata associated with series.
    I.e. Law & Order.
    """
    pass



from django import forms
from django.utils.safestring import mark_safe

class RadioFieldImageRenderer(forms.widgets.RadioFieldRenderer):
    def render(self):
        """
        Outputs a <ul> for this set of radio fields.
        """
        output = []
        output.append('<table style="border: 0 solid black"><tr>')
        for choice in self.choices:
            output.append("""
<th style="border: 0 solid black">
<center>
<input type="radio" name="%s" value="%s" id="%s" >
%s
</center>
<br />
<img src="%s" height="130"/></th>""" % (self.name, choice['file_path'], self.attrs['id'], choice['label'], choice['media_path']))
        output.append('</tr></table>')
        return mark_safe(u''.join(output))



import unittest

from tube.models import ClipBase
class ClipBaseTestCase(unittest.TestCase):
    def test_duration_as_hours_minute_seconds(self):
        
        # 0 duration. All should be 0.
        clip_base = ClipBase(duration=0)
        result = clip_base.duration_as_hours_minute_seconds
        self.failUnless(result['hours']==0, result)
        self.failUnless(result['minutes']==0, result)
        self.failUnless(result['seconds']==0, result)
        
        # Random tests. XXX: Flesh out with considered cases.
        clip_base = ClipBase(duration=3660)
        result = clip_base.duration_as_hours_minute_seconds
        self.failUnless(result['hours']==1, result)
        self.failUnless(result['minutes']==1, result)
        self.failUnless(result['seconds']==0, result)
        
        clip_base = ClipBase(duration=3665)
        result = clip_base.duration_as_hours_minute_seconds
        self.failUnless(result['hours']==1, result)
        self.failUnless(result['minutes']==1, result)
        self.failUnless(result['seconds']==5, result)
        
        clip_base = ClipBase(duration=74564376)
        result = clip_base.duration_as_hours_minute_seconds
        self.failUnless(result['hours']==20712, result)
        self.failUnless(result['minutes']==19, result)
        self.failUnless(result['seconds']==36, result)



import math
import mimetypes
import os
import subprocess
import tempfile

from django import forms
from django.contrib import admin
from django.contrib.contenttypes.models import ContentType
from django.conf import settings
from django.core.files.uploadedfile import InMemoryUploadedFile

from panya.admin import ModelBaseAdmin
from tube.models import AgeRestriction, Channel, Clip, Series
from tube.widgets import RadioFieldImageRenderer

import pyffmpeg

class ChannelAdmin(ModelBaseAdmin):
    def queryset(self, request):
        """
        Limit queryset to Channel objects (those of content type Channel)
        """
        channel_type = ContentType.objects.get_for_model(Channel)
        return self.model.objects.filter(content_type=channel_type)

class ClipForm(forms.ModelForm):
    image_choice = forms.fields.CharField(
        widget=forms.RadioSelect(
            renderer=RadioFieldImageRenderer
        ),
        required=False,
        help_text='Upload a media resource above to enable scraped image selection.'
    )
    
    class Meta:
        model = Clip

    def __init__(self, *args, **kwargs):
        super(ClipForm, self).__init__(*args, **kwargs)
       
        # Generate choices from scraped thumbnails.
        if not self.is_bound or self.errors:
            choices = []
            instance = self.instance
            if instance.media_resource:
                thumb_paths = self.generate_thumbs(instance.media_resource)

                if thumb_paths:
                    for idx, val in enumerate(thumb_paths):
                        choices.append({'file_path': val['file_path'], 'media_path': val['media_path'], 'label': 'Scraped Image %s' % (idx + 1)})
        
                    self.fields['image_choice'].widget.choices = choices
                    self.fields['image_choice'].help_text='Select a scraped image you want to use for this clip. Alternatively upload a custom image below.'

    def generate_thumbs(self, resource):
        """
        Generate image thumbnails from provided media resource.
        """
        # Open video stream.
        stream = pyffmpeg.VideoStream()
        try:
            stream.open(resource.path)
        except:
            stream.close()
            # Fail silently if something goes wrong during stream opening.
            return None

        # Calculate frame indexes from which to generate thumbs.
        fps = stream.tv.get_fps()
        duration = stream.vr.duration_time()
        frame_count = int(math.floor(duration) * math.floor(fps))
        frame_indexes = range(0, frame_count, frame_count / 4)[:4]

        # Offset first frame by 3 seconds.
        frame_indexes[0] = int(fps) * 3

        paths = []
        for index in frame_indexes:
            # Get image from stream.
            try:
                image = stream.GetFrameNo(index)
                file_name = '%s.thumb%s.png' % (resource.name.split('/')[-1], index)
                file_path = '%s%s' % (settings.MEDIA_ROOT, file_name)
                image.save(file_path, 'PNG')
                paths.append({'file_path': file_path, 'media_path': '%s%s' % (settings.MEDIA_URL, file_name)})
            except IOError, e:
                # Abort if we can not scrape images. most likely means something is wrong with the clip.
                return None

        stream.close()
        return paths

    def clean_media_resource(self):
        """
        Validate provided media format against ffmpeg by trying to grab a frame from it.
        Automatically assume format is invalid if it doesn't have a video mimetype.
        """
        error_msg = "Unsupported media format."
        data = self.cleaned_data['media_resource']
        
        # Only validate on change.
        if 'media_resource' not in self.changed_data:
            return data

        # Check for video mimetype.
        if 'video' not in data.content_type:
            raise forms.ValidationError(error_msg)
            
        # Create temp file and dump uploaded data to it.
        temp_filename = tempfile.mkstemp("-validate-%s" % data.name)[1]
        temp_file = open(temp_filename, 'wb')
        for chunk in data.chunks():
            temp_file.write(chunk)
        temp_file.close()

        # Try to grab a frame through ffmpeg and record the process result.
        temp_image_filename = '%s.png' % temp_filename
        result = subprocess.call(['ffmpeg', '-vframes', '1', '-i', temp_filename, temp_image_filename])

        # Remove temp files.
        os.remove(temp_filename)
        os.remove(temp_image_filename)

        # If ffmpeg failed assume the format is invalid
        if result != 0:
            raise forms.ValidationError(error_msg)
        else:
            return data
        
class ClipAdmin(ModelBaseAdmin):
    form = ClipForm
    
    def __init__(self, model, admin_site):
        from copy import deepcopy
        super(ClipAdmin, self).__init__(model, admin_site)
        fieldsets = deepcopy(self.fieldsets)
        for fieldset in fieldsets:
            if fieldset[0] == 'Image':
                fieldset[1]['fields'] = ('image_choice',) + fieldset[1]['fields']

        self.fieldsets = fieldsets
    
    def assign_image_to_field(self, file_path, field):
        f = open(file_path, 'r')
        
        # guess content type from filename
        content_type = mimetypes.guess_type(file_path)[0]
        # calc size
        size = len(f.read())

        # create InMemoryUploadedFile
        field_file = InMemoryUploadedFile(f, field.field.name, file_path, content_type, size, None)
        # save field file to field
        field.save(field_file.name, field_file)

        f.close()
    
    def save_model(self, request, obj, form, change):
        """
        Given a model instance save it to the database.
        """
        if 'image_choice' in form.changed_data:
            file_path = form.cleaned_data['image_choice']
            self.assign_image_to_field(file_path, obj.image)
        
        obj.save()

admin.site.register(AgeRestriction)
admin.site.register(Channel, ChannelAdmin)
admin.site.register(Clip, ClipAdmin)
admin.site.register(Series, ModelBaseAdmin)
